import os
from typing import List
from matplotlib.figure import Figure
from pandas import DataFrame
from serapis.event import Event, Catalog, EventData


def test_create_event_instance(event_instance_attrs: List[str]):
    event_obj = Event(f"test")
    assert event_obj.depth_prefix == "DepthMax"
    assert isinstance(event_obj.reference_index, DataFrame)
    assert all([i in event_obj.__dict__.keys() for i in event_instance_attrs])


def test_create_from_overtopping(overtopping_file: str):
    event_obg = Event("test")
    event_obg.create_from_overtopping(overtopping_file)
    assert hasattr(event_obg, "event_index")
    assert isinstance(event_obg.event_index, DataFrame)


def test_calculateVolumeError(
    overtopping_file: str, volume_error_file: str, event_index_volume_attrs: List[str]
):
    event_obg = Event("test")
    event_obg.create_from_overtopping(overtopping_file)
    event_obg.calculate_volume_error(volume_error_file)
    assert all([i in event_obg.event_index.columns for i in event_index_volume_attrs])


def test_read_event_index(
    event_index_file: str, volume_error_file: str, event_index_volume_attrs2: List[str]
):
    event = Event.read_event_index("test", event_index_file, start="1955-01-01")
    assert isinstance(event.event_index, DataFrame)
    assert all([i in event.event_index.columns for i in event_index_volume_attrs2])


def test_get_event_start(event_index_file: str, event_index_volume_attrs2: List[str]):
    event = Event.read_event_index("test", event_index_file, start="1955-01-01")
    max_overtopping_ind = event.event_index["cells"].idxmax()
    start_ind, start_day = event.get_event_start(max_overtopping_ind)
    end_ind, end_day = event.get_event_end(max_overtopping_ind)
    assert start_day == 8195
    assert end_day == 8200


def test_get_event_by_index(
    event_index_file: str, volume_error_file: str, event_index_volume_attrs2: List[str]
):
    event = Event.read_event_index("test", event_index_file, start="1955-01-01")
    event_details = event.get_event_by_index(1)
    assert event_details == {"start": 35, "end": 38, "cells": 1023}


def test_detailed_overtopping(
    event_index_file: str,
):
    event = Event.read_event_index("test", event_index_file, start="1955-01-01")
    event.two_d_result_path = "tests/data"
    fig, ax, opts = event.histogram(8200, 0, 1)
    # {"n": n, "bins": bins, "patches": patches}
    assert isinstance(fig, Figure)
    assert all(elem in opts.keys() for elem in ["n", "bins", "patches"])


class TestCatalog:
    def test_read_file(self):
        path = "tests/data/catalog.yaml"
        catalog = Catalog.read_file(path)
        assert isinstance(catalog.catalog, dict)
        assert catalog.events == [19, 50]

    def test_to_file(self):
        path = "tests/data/catalog.yaml"
        save_to = "tests/data/save_catalog.yaml"
        if os.path.exists(save_to):
            os.remove(save_to)

        catalog = Catalog.read_file(path)
        catalog.to_file(save_to)
        assert os.path.exists(path)
        os.remove(save_to)

    def test_dunder_methods(self):
        path = "tests/data/catalog.yaml"
        catalog = Catalog.read_file(path)
        # test __len__
        assert len(catalog) == 2
        # test __iter__
        assert len(list(catalog)) == 2
        # test __getitem__
        f = catalog[19]
        assert isinstance(catalog[19], EventData)
        assert catalog[19].data == {
            "day": 19,
            "depth": {
                5: [0.0199, 0.0099, 0.14, 0.09, 0.0199, 0.0399],
                29: [0.4199, 0.31, 0.31, 0.3499, 0.34, 0.4099],
            },
            "reaches": [29.0, 5.0],
        }
        # test __setitem__
        catalog[20] = {}
        assert 20 in catalog.events

    # def test__getitem__(self):
    #     path = "tests/data/catalog.yaml"
    #     catalog = Catalog.read_file(path)
    #     # test __getitem__
    #     assert isinstance(catalog[19], EventData)
    #     assert catalog[19].data == {
    #         "day": 19,
    #         "depth": {
    #             5: [0.0199, 0.0099, 0.14, 0.09, 0.0199, 0.0399],
    #             29: [0.4199, 0.31, 0.31, 0.3499, 0.34, 0.4099],
    #         },
    #         "reaches": [29.0, 5.0],
    #     }
    #     # test __setitem__
    #     catalog[20] = {}
    #     assert 20 in catalog.events




"""Inputs."""
import datetime as dt
import os
import zipfile
from typing import Union
import numpy as np
import pandas as pd
from loguru import logger
from osgeo import gdal
from statista.distributions import Distributions
from statista.eva import ams_analysis
from serapis.river import River


class Inputs(River):
    """Hydraulic model Inputs class.

    Methods
    -------
        1- ExtractHydrologicalInputs
        2- StatisticalProperties
        3- WriteHQFile
        4- ReturnPeriod
        5- ReadRIMResult
        6- CreateTraceALL
    """

    def __init__(self, name, version: int = 3):
        """Input.

            Inputs is an object to create the inputs for the river model

        Parameters
        ----------
        name: [str]

        version: [int]
            model version
        """
        self.Name = name
        self.version = version
        self.statistical_properties = None
        self.distribution_properties = None

    def extract_hydrological_inputs(
        self, weather_generator, file_prefix, realization, path, locations, save_path
    ):
        """extract_hydrological_inputs.

        Parameters
        ----------
        weather_generator : TYPE
            DESCRIPTION.
        file_prefix : TYPE
            DESCRIPTION.
        realization : [Integer]
            type the number of the realization (the order of the 1000-year run by rainfall runoff model).
        path : [String]
             rrm_result_file is the naming format you used in naming the result files of the discharge values stored
             with the name of the file as out+realization number + .dat (ex out15.dat).
        locations : [String]
            text file containing the list of sub-basins IDs or computational nodes ID you have used to run the rrm
            and store the results.
        save_path : [String]
            path to the folder where you want to save the separate file for each sub-basin.

        Returns
        -------
        None.
        """
        if weather_generator:
            # weather_generator
            rrm_result_file = file_prefix + str(realization) + ".dat"
            # 4-5
            # check whether the name of the realization the same as the name of 3 the saving file or not to prevent
            # any confusion in saving the files
            if int(realization) <= 9:
                assert int(rrm_result_file[-5:-4]) == int(
                    save_path[-1]
                ), " Wrong files sync "
            else:
                assert int(rrm_result_file[-6:-4]) == int(
                    save_path[-2:]
                ), " Wrong files sync "
        else:
            # Observed data.
            rrm_result_file = file_prefix + str(realization) + ".dat"

        # rrm writes the year as the first column then day as a second column and the discharge values start from the
        # thirst column, so you have to write number of columns to be ignored at the beginning

        ignore_columns = 2

        # read SWIM result file
        swim_data = pd.read_csv(
            f"{path}/{rrm_result_file}", delimiter=r"\s+", header=None
        )
        nodes = pd.read_csv(f"{path}/{locations}", header=None)

        for i in range(len(nodes)):
            swim_data.loc[:, i + ignore_columns].to_csv(
                f"{save_path}/{nodes.loc[i, 0]}.txt", header=None, index=None
            )

    def get_statistical_properties(
        self,
        gauges: list,
        rdir: str,
        start: str,
        warm_up_period: int,
        save_plots: bool,
        save_to: str,
        filter_out: Union[bool, float, int] = False,
        distribution: str = "GEV",
        method: str = "lmoments",
        quartile: float = 0,
        significance_level: float = 0.1,
        file_extension: str = ".txt",
        date_format: str = "%Y-%m-%d",
    ):
        """get_statistical_properties.

            get_statistical_properties method reads the discharge hydrographs of a rainfall runoff model and
            hydraulic model for some computational nodes and calculates some statistical properties.

            the code assumes that the time series are of a daily temporal resolution, and that the hydrological year
            is 1-Nov/31-Oct (Petrow and Merz, 2009, JoH).

        Parameters
        ----------
        gauges : [list]
            The list which contains the ID of the gauges you want to do the statistical analysis for, the ObservedFile
            should contain the discharge time series of these nodes in order.
        rdir : [str]
            The directory where the time series files exist.
        start : [string]
            the beginning date of the time series.
        warm_up_period : [int]
            The number of days you want to neglect at the beginning of the simulation (warm up period).
        save_plots : [Bool]
            True if you want to save the plots.
        save_to : [str]
            the rdir where you want to save the statistical properties.
        filter_out: [Bool]
            For observed or hydraulic model data it has gaps at times, where the model did not run or gaps in the
            observed data if these gap days are filled with a specific value, and you want to ignore it here give
            filter_out = Value you want.
        distribution: [str]
            Default is "GEV".
        method: [str]
            available methods are 'mle', 'mm', 'lmoments', optimization. Default is "lmoments".
        quartile: [float]
            Default is 0.
        significance_level:
            Default is [0.1].
        file_extension: [str]
            Default is '.txt'.
        date_format: [str]
            Default is "%Y-%m-%d".

        Returns
        -------
        statistical-properties.csv:
            file containing some statistical properties like mean, std, min, 5%, 25%, median, 75%, 95%, max, t_beg,
            t_end, nyr, q1.5, q2, q5, q10, q25, q50, q100, q200, q500.
        """
        if not isinstance(gauges, list):
            raise TypeError("gauges should be either a rdir or a list")

        # hydrographs
        time_series = pd.DataFrame()
        # for the hydraulic model results
        logger.info(
            "The function ignores the date column in the time series files and starts from the given start"
            "parameter to the function so check if it is the same start date as in the files"
        )

        skip = []
        for gauge_i in gauges:
            try:
                time_series.loc[:, int(gauge_i)] = pd.read_csv(
                    f"{rdir}/{int(gauge_i)}{file_extension}",
                    skiprows=1,
                    header=None,
                )[1].tolist()
            except FileNotFoundError:
                logger.warning(f"File {int(gauge_i)}{file_extension} does not exist")
                skip.append(int(gauge_i))

        start_date = dt.datetime.strptime(start, date_format)
        end_date = start_date + dt.timedelta(days=time_series.shape[0] - 1)
        ind = pd.date_range(start_date, end_date)
        time_series.index = ind

        # neglect the first year (warmup year) in the time series
        time_series = time_series.loc[
            start_date + dt.timedelta(days=warm_up_period) : end_date, :
        ]

        statistical_properties, distribution_properties = ams_analysis(
            time_series,
            ams_start="A-OCT",
            save_plots=save_plots,
            save_to=save_to,
            filter_out=filter_out,
            distribution=distribution,
            method=method,
            # estimate_parameters=estimate_parameters,
            quartile=quartile,
            alpha=significance_level,
        )

        # Output file
        statistical_properties.to_csv(
            f"{save_to}/statistical-properties.csv", float_format="%.4f"
        )

        distribution_properties.to_csv(
            f"{save_to}/DistributionProperties.csv", float_format="%.4f"
        )

        self.statistical_properties = statistical_properties
        self.distribution_properties = distribution_properties

    @staticmethod
    def StringSpace(Inp):
        """StringSpace."""
        return str(Inp) + "  "

    def return_period(
        self,
        maps_path,
        prefix,
        distribution_properties_file,
        trace_file,
        sub_basin_file,
        replacement_file,
        hydrological_inputs_path,
        sub_id_map,
        extra_subs_file,
        from_file,
        to_file,
        save_to,
        wpath,
    ):
        """Return period."""
        all_results = os.listdir(maps_path)
        # list of the Max Depth files only
        max_depth_list = list()
        for i in range(len(all_results)):
            if all_results[i].startswith(prefix):
                max_depth_list.append(all_results[i])
        # Read Inputs
        # read the Distribution parameters for each upstream computational node
        distribution_properties = pd.read_csv(distribution_properties_file)
        us_node = pd.read_csv(trace_file, header=None)
        us_node.columns = ["SubID", "US", "DS"]
        # get the sub-basin id from the guide file it is the same shape in RIM1.0 and RIM2.0
        subs_id = pd.read_csv(sub_basin_file, header=None, usecols=[0])

        replacement_sub = pd.read_csv(replacement_file)

        # read the hydrograph for all the US nodes
        # start_date = "1950-1-1"
        # start_date = dt.datetime.strptime(start_date,"%Y-%m-%d")
        # ind = pd.date_range(start_date, StartDate + dt.timedelta(days = NoYears*365), freq = "D")

        ind = range(
            1,
            len(
                pd.read_csv(
                    hydrological_inputs_path
                    + "/"
                    + str(int(us_node.loc[subs_id.loc[10, 0] - 1, "US"]))
                    + ".txt"
                ).values
            ),
        )

        hydrographs = pd.DataFrame(index=ind, columns=subs_id[0].to_list())

        for i in range(len(subs_id)):
            #    i=1
            # search for the SubId in the us_node, or it is listed by order, so subID=343 exist
            # in the row 342 (SubID-1)
            # np.where(us_node['SubID'] == subs_id.loc[i,0])
            try:
                if int(us_node.loc[subs_id.loc[i, 0] - 1, "US"]) != -1:
                    hydrographs.loc[:, subs_id.loc[i, 0]] = pd.read_csv(
                        hydrological_inputs_path
                        + "/"
                        + str(int(us_node.loc[subs_id.loc[i, 0] - 1, "US"]))
                        + ".txt"
                    ).values[: len(hydrographs)]
            except:
                other_sub_loc = np.where(
                    replacement_sub["missing"] == subs_id.loc[i, 0]
                )[0][0]
                if (
                    int(
                        us_node.loc[
                            replacement_sub.loc[other_sub_loc, "replacement"] - 1, "US"
                        ]
                    )
                    != -1
                ):
                    hydrographs.loc[:, subs_id.loc[i, 0]] = pd.read_csv(
                        hydrological_inputs_path
                        + "/"
                        + str(
                            int(
                                us_node.loc[
                                    replacement_sub.loc[other_sub_loc, "replacement"]
                                    - 1,
                                    "US",
                                ]
                            )
                        )
                        + ".txt"
                    ).values[: len(hydrographs)]

        # read sub basin map id
        sub_id_map = gdal.Open(sub_id_map)
        sub_id_map_v = sub_id_map.ReadAsArray()

        # read the added subs' reference text file
        extra_subs = pd.read_csv(extra_subs_file)

        # function to write the numbers in the ASCII file

        # read Max depth map
        check = list()
        klist = list()

        if to_file == "end" or to_file > len(max_depth_list):
            to_file = len(max_depth_list)

        # from_file = 48
        # to_file = from_file +1

        for k in range(from_file, to_file):
            try:
                # open the zip file
                compressed_file = zipfile.ZipFile(f"{maps_path}/{max_depth_list[k]}")
            except:
                print("Error Opening the compressed file")
                check.append(max_depth_list[k][len(prefix) : -4])
                klist.append(k)
                continue

            # get the file name
            file_name = compressed_file.infolist()[0]
            # get the time step from the file name
            timestep = int(file_name.filename[len(prefix) : -4])
            print("File= " + str(timestep))

            asci_file = compressed_file.open(file_name)
            f = asci_file.readlines()
            spatial_ref = f[:6]
            ascii_raw = f[6:]
            # asci_file = compressed_file.open(file_name)
            # ascii_raw = asci_file.readlines()[6:]
            rows = len(ascii_raw)
            cols = len(ascii_raw[0].split())
            max_depth = np.ones((rows, cols), dtype=np.float32)
            # read the ascii file
            for i in range(rows):
                x = ascii_raw[i].split()
                max_depth[i, :] = list(map(float, x))

            # check on the values of the water depth
            #    if np.shape(max_depth[np.isnan(max_depth)])[0] > 0:
            #        check.append(timestep)
            #        print("Error Check Max Depth values")
            #        continue

            # plotting to check values
            #    fromrow = np.where(max_depth == max_depth.max())[0][0]
            #    fromcol = np.where(max_depth == max_depth.max())[1][0]
            #    plt.imshow(max_depth[fromrow-20:fromrow+20,fromcol-20:fromcol+20])
            #    plt.imshow(max_depth)
            #    plt.colorbar()

            # get the Peak of the hydrograph for the whole event
            # (14 days before the end of the event)
            max_values_file = hydrographs.loc[timestep - 14 : timestep, :]
            max_values = max_values_file.max().values.tolist()
            return_period = list()

            # Calculate the Return period for the max Q at this time step for each
            for i in range(len(max_values)):
                # if the sub-basin is a lateral and not routed in RIM, it will not have a
                # hydrograph
                if np.isnan(max_values[i]):
                    return_period.append(np.nan)
                if not np.isnan(max_values[i]):
                    # np.where(us_node['SubID'] == subs_id.loc[i,0])
                    try:
                        downstream_node = us_node.loc[subs_id.loc[i, 0] - 1, "US"]
                        loc = np.where(
                            distribution_properties["id"] == downstream_node
                        )[0][0]
                    except IndexError:
                        other_sub_loc = np.where(
                            replacement_sub["missing"] == subs_id.loc[i, 0]
                        )[0][0]
                        downstream_node = us_node.loc[
                            replacement_sub.loc[other_sub_loc, "replacement"] - 1, "US"
                        ]
                        loc = np.where(
                            distribution_properties["id"] == downstream_node
                        )[0][0]

                    # to get the Non-Exceedance probability for a specific Value.
                    parameters = dict(
                        loc=distribution_properties.loc[loc, "loc"],
                        scale=distribution_properties.loc[loc, "scale"],
                    )
                    dist = Distributions("Gumbel", max_values[i])
                    rp = dist.get_rp(parameters)

                    return_period.append(round(rp, 2))

            try:
                retun_period_map = np.ones((rows, cols), dtype=np.float32) * 0
                for i in range(rows):
                    for j in range(cols):
                        # print("i = " + str(i) + ", j= " + str(j))
                        if not np.isnan(max_depth[i, j]):
                            if max_depth[i, j] > 0:
                                # print("i = " + str(i) + ", j= " + str(j))
                                # if the sub basin is in the Sub ID list
                                if sub_id_map_v[i, j] in subs_id[0].tolist():
                                    # print("Sub = " + str(sub_id_map_v[i,j]))
                                    # go get the return period directly
                                    retun_period_map[i, j] = return_period[
                                        np.where(subs_id[0] == sub_id_map_v[i, j])[0][0]
                                    ]
                                else:
                                    # print("Extra  Sub = " + str(sub_id_map_v[i,j]))
                                    # the sub ID is one of the added subs not routed by RIM, so it existed in the
                                    # extra_subs list with a reference to a SubID routed by RIM.
                                    rim_sub = extra_subs.loc[
                                        np.where(
                                            extra_subs["addSub"] == sub_id_map_v[i, j]
                                        )[0][0],
                                        "rim_sub",
                                    ]
                                    retun_period_map[i, j] = return_period[
                                        np.where(subs_id[0] == rim_sub)[0][0]
                                    ]
            except:
                print("Error")
                check.append(timestep)
                klist.append(k)
                continue

            # save the return period ASCII file
            file_name = "ReturnPeriod" + str(timestep) + ".asc"

            with open(save_to + "/" + file_name, "w") as File:
                # write the first lines
                for i in range(len(spatial_ref)):
                    File.write(str(spatial_ref[i].decode()[:-2]) + "\n")

                for i in range(np.shape(retun_period_map)[0]):
                    File.writelines(list(map(self.StringSpace, retun_period_map[i, :])))
                    File.write("\n")

            # zip the file
            with zipfile.ZipFile(
                save_to + "/" + file_name[:-4] + ".zip", "w", zipfile.ZIP_DEFLATED
            ) as newzip:
                newzip.write(save_to + "/" + file_name, arcname=file_name)
            # delete the file
            os.remove(save_to + "/" + file_name)

        check = list(zip(check, klist))
        if len(check) > 0:
            np.savetxt(wpath + "CheckWaterDepth.txt", check, fmt="%6d")

    def create_trace_all(
        self,
        config_file_path,
        rim_subs_file,
        trace_file,
        us_only: int = 1,
        hydrological_inputs_file: str = "",
    ):
        """CreateTraceALL.

        Parameters
        ----------
        config_file_path: [String]
            SWIM configuration file.
        rim_subs_file: [String]
            path to text file with all the ID of SWIM sub-basins.
        trace_file: TYPE
            DESCRIPTION.
        us_only: TYPE, optional
            DESCRIPTION. The default is 1.
        hydrological_inputs_file: TYPE, optional
            DESCRIPTION. The default is ''.

        Returns
        -------
        None.
        """
        # reading the file
        config = pd.read_csv(config_file_path, header=None)
        # process the Configuration file
        # get the route rows from the file
        route = pd.DataFrame(columns=["No", "DSnode", "SubID", "USnode", "No2"])

        j = 0
        for i in range(len(config)):
            if config[0][i].split()[0] == "route":
                route.loc[j, :] = list(map(int, config[0][i].split()[1:]))
                j = j + 1

        # get RIM Sub-basins
        subs = pd.read_csv(rim_subs_file, header=None)
        subs = subs.rename(columns={0: "SubID"})

        subs["US"] = None
        subs["DS"] = None

        for i in range(len(subs)):
            try:
                # if the sub-basin is in the route array, so it is routed by rrm.
                loc = np.where(route["SubID"] == subs.loc[i, "SubID"])[0][0]
                subs.loc[i, "US"] = int(route.loc[loc, "USnode"])
                subs.loc[i, "DS"] = int(route.loc[loc, "DSnode"])
            except IndexError:
                # if the sub-basin is not in the route array, so it is not routed by rrm.
                # but still can be routed using RIM

                # subs.loc[i,'US'] = None
                # subs.loc[i,'DS'] = None
                subs.loc[i, "US"] = -1
                subs.loc[i, "DS"] = -1

        # Save the file with the same format required for the hg R code

        subs.to_csv(trace_file, index=None, header=True)
        #    ToSave = subs.loc[:,['SubID','US']]
        #    ToSave['Extra column 1'] = -1
        #    ToSave['Extra column 2'] = -1
        #    ToSave.to_csv(save_path + trace_file,header = None, index = None)



"""1D riven Events."""
import os
from typing import Iterator, Tuple, Any, Union, Dict, List
from pathlib import Path
import datetime as dt
import yaml
from dataclasses import dataclass
import numpy as np
import pandas as pd
from pandas.core.frame import DataFrame
from pyramids.dataset import Dataset
from cleopatra.statistical_glyph import StatisticalGlyph as Statistic


class Catalog:
    """Flood catalog.

    - the catalog is a dictionary
        - keys can be a string or a numeric.
        - values can only be
            - numeric
            - lists:
                list of numerics
            - nested dictionary (so this inner dictionary can have only lists or numeric as values)

    - So the following is how the Catalog is structured, where we have two events 19, and 25, for each event
    there is some properties that we want to store for this event
    19:
    {
        day: 19
        depth: {5: [0.0199, 0.0099, 0.1400], 29: [0.4199, 0.3100, 0.3100]}
        reaches: [29.0, 5.0]
      }
    25:
    {
        day: 25
        depth: {3: [0.0199, 0.0099, 0.1400], 15: [0.4199, 0.3100, 0.3100]}
        reaches: [3.0, 15.0]
        }

    - the catalog will be saved to disk in a yaml file with the following format.
    19:
      day: 19
      depth:
        5:
        - 0.0199
        - 0.0099
        - 0.1400
        29:
        - 0.4199
        - 0.3100
        - 0.3100
      reaches:
      - 29.0
      - 5.0
    25:
      day: 25
      depth:
        3:
        - 0.0199
        - 0.0099
        - 0.1400
        15:
        - 0.4199
        - 0.3100
        - 0.3100
      reaches:
      - 3.0
      - 15.0
    """

    float_3 = lambda x: float(round(x, 3))

    def __init__(self, catalog: dict):
        self._catalog = catalog

    @property
    def catalog(self) -> Dict:
        """Catalog."""
        return self._catalog

    @property
    def events(self) -> List:
        """Event days."""
        return list(self._catalog.keys())

    def __str__(self):
        """Print event details."""
        message = f"""
                    Number of Events: {len(self)}
                    Events: {self.events}
                """
        return message

    def __repr__(self):
        """Print event details."""
        message = f"""
                    Number of Events: {len(self)}
                    Events: {self.events}
                """
        return message

    def __len__(self):
        """Length."""
        return len(self._catalog.keys())

    def __iter__(self) -> Iterator[Any]:
        """Iterate."""
        return iter(self._catalog.items())

    def __setitem__(self, key, value):
        """Set event."""
        if key in self.events:
            raise KeyError(f"event: {key} already exists")
        self._catalog.update({key: value})

    def __getitem__(self, item):
        """Get Event."""
        if item not in self.events:
            raise KeyError(
                f"there is no event: {item}, available events are: {self.events}"
            )
        data = self._catalog[item]
        return EventData(data)

    @classmethod
    def read_file(cls, path: str):
        """Read catalog.

        Parameters
        ----------
        path: str
        """
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(
                f"the path you entered does not exist: {path.absolute()}"
            )

        with open(path.absolute(), "r") as stream:
            catalog = yaml.safe_load(stream)

        return cls(catalog)

    @staticmethod
    def _serialize(in_dict: dict):
        # convert the keys to int
        new_dict = {}
        for key, val in in_dict.items():
            try:
                key = int(key)
            except ValueError:
                pass

            if isinstance(val, list):
                new_dict[key] = list(map(float, in_dict[key]))  # Catalog.float_3
            elif isinstance(val, dict):
                new_dict[key] = Catalog._serialize(val)
            else:
                new_dict[key] = int(val) if val is not None else None

        return new_dict

    def to_file(self, path: str):
        """Save catalog.

        Parameters
        ----------
        path: path
        """
        cat = self._serialize(self.catalog)

        with open(path, "w") as outfile:
            yaml.dump(cat, outfile, default_flow_style=False)


@dataclass
class EventData:
    """Event Data in the catalog."""

    data: Dict

    def parse(self):
        """Parse event data."""
        self.day = self.data.get("day")
        self.depth = self.data.get("depth")
        self.reaches = self.data.get("reaches")
        self.overtopping = self._get_overtopping()

    # def __post_init__(self):
    #     """Parse event data."""
    #     self.day = self.data.get("day")
    #     self.depth = self.data.get("depth")
    #     self.reaches = self.data.get("reaches")
    #     self.overtopping = self._get_overtopping()

    def _get_overtopping(self):
        over_top = self.data.get("overtopping")
        df = pd.DataFrame(columns=["cross-sections", "overtopping"])
        df["cross-sections"] = list(map(int, over_top["cross-sections"]))
        df["overtopping"] = over_top["volume"]
        return df


class Event:
    """Event.

        The Event class reads all the results of the Hydraulic model to preform all kind of analysis on flood event
        basis and the overtopping.

    Methods
    -------
    indexToDate
    createEventIndex
    getAllEvents
    readOvertopping
    calculateVolumeError
    OverlayMaps
    readEventIndex
    histogram
    drop
    to_csv
    getEventBeginning
    getEventEnd
    prepareForPlotting
    listAttributes
    """

    def __init__(
        self,
        name: str,
        start: Union[str, dt.datetime] = "1950-1-1",
        event_index: DataFrame = None,
        left_overtopping_suffix: str = "_left.txt",
        right_overtopping_suffix: str = "_right.txt",
        depth_prefix: str = "DepthMax",
        duration_prefix: str = "Duration",
        return_period_prefix: str = "ReturnPeriod",
        compressed: bool = True,
    ):
        """Event. To instantiate the Event class you need to provide the following arguments.

        Parameters
        ----------
        name : [str]
            DESCRIPTION.
        start : [str], optional
            start date. The default is "1950-1-1".
        left_overtopping_suffix : [str], optional
            the prefix you used to name the overtopping form the left bank files.
            The default is "_left.txt".
        right_overtopping_suffix : TYPE, optional
            the prefix you used to name the overtopping form the right bank files.
            The default is "_right.txt".
        depth_prefix : [str], optional
            the prefix you used to name the Max depth raster result maps.
            The default is "DepthMax".
        duration_prefix : [str], optional
            the prefix you used to name the inundation duration raster result maps.
            The default is "Duration".
        return_period_prefix : [str], optional
            the prefix you used to name the Return Period raster result maps.
            The default is "ReturnPeriod".
        compressed : [bool], optional
            True if the result raster/ascii files are compressed. The default is True.

        Returns
        -------
        None.
        """
        # instance attribute
        self.name = name
        self.start = (
            start
            if not isinstance(start, str)
            else dt.datetime.strptime(start, "%Y-%m-%d")
        )
        self.reference_index = self.create_reference_index(self.start, freq="D")
        self._event_index = event_index
        if event_index is not None:
            self._number_events = len(self.event_index.loc[:, "index"].unique())

        self._left_overtopping_suffix = left_overtopping_suffix
        self._right_overtopping_suffix = right_overtopping_suffix
        self._depth_prefix = depth_prefix
        self._duration_prefix = duration_prefix
        self._return_period_prefix = return_period_prefix
        self._two_d_result_path = ""
        self.compressed = compressed

        # create dictionary to store any extracted values from maps
        self.extracted_values = dict()
        self.event_beginning = None
        self.end_days = None

    @staticmethod
    def create_reference_index(start: dt.datetime, days: int = 36890, freq: str = "D"):
        """Create Event Index."""
        end = start + dt.timedelta(days=days)
        Ref_ind = pd.date_range(start, end, freq=freq)
        # the last day is not in the results day Ref_ind[-1]
        # write the number of days + 1 as python does not include the last number in the range
        # 19723 days so write 19724
        reference_index = pd.DataFrame(index=list(range(1, days + 1)))
        reference_index["date"] = Ref_ind[:-1]

        return reference_index

    @property
    def event_index(self):
        """Event index."""
        return self._event_index

    @property
    def numer_events(self):
        """Events Number."""
        return self._number_events

    @property
    def depth_prefix(self):
        """Depth Prefix."""
        return self._depth_prefix

    @depth_prefix.setter
    def depth_prefix(self, value):
        self._depth_prefix = value

    @property
    def duration_prefix(self):
        """Duration Prefix."""
        return self._duration_prefix

    @duration_prefix.setter
    def duration_prefix(self, value):
        self._duration_prefix = value

    @property
    def two_d_result_path(self):
        """2D result path."""
        return self._two_d_result_path

    @two_d_result_path.setter
    def two_d_result_path(self, value):
        self._two_d_result_path = value

    @staticmethod
    def ordinal_to_date(df: DataFrame, reference_index: DataFrame):
        """IndexToDate. get the date coresponding to a given index.

        Returns
        -------
        Date
        """
        # convert the index into date
        dateFn = lambda i: reference_index.loc[i, "date"]
        # get the date the column 'id'
        date = df.loc[:, "id"].to_frame().applymap(dateFn)
        df["date"] = date
        return df

    def get_event_by_index(self, event_i: int) -> Dict[str, str]:
        """get_event_by_index.

            using the index get the event last day.

        Parameters
        ----------
        event_i: [int]
            the order of the event.

        Returns
        -------
        Dict:
            {'start': 35, 'end': 38, 'cells': 1023}
        """
        if event_i > self.numer_events:
            raise ValueError(f"There are only {self.numer_events}, given: {event_i}")
        # ind = self.event_index.loc[:, "index"].values
        # row = np.where(ind == event_i)[0][-1]
        # return self.event_index.loc[row, "id"]
        df = self.event_index[self.event_index["index"] == event_i].reset_index(
            drop=True
        )
        ind = len(df) - 1
        event_data = {
            "start": df.loc[0, "id"],
            "end": df.loc[ind, "id"],
        }
        if "cells" in df.columns:
            event_data["cells"] = df.loc[ind, "cells"]

        return event_data

    @classmethod
    def create_from_index(
        cls,
        name: str,
        path: str,
        start: str,
        freq: str = "D",
        fmt: str = "%Y-%m-%d",
        event_duration: int = 10,
    ):
        """create_from_index.

            create_from_index takes the path to the index file result from the 2D model and creates a data frame to
            start adding the components of the event_index table.

        Parameters
        ----------
        name: [str]
            name of the river.
        path: [str]
            path of the index file result from the 2D model.
        start: [str]
            start date of the simulation.
        freq: [str]
            temporal resolution of the simulation. Default is "D".
        fmt: [str]
            Default is "%Y-%m-%d".
        event_duration: [int]
            Default is 10.

        Returns
        -------
        event_index: [dataframe]
            this method creates an instance attribute of type dataframe with columns ['id','continue', 'ind_diff',
            'duration']
        """
        # read the index file (containing the id of the days, where flood happens (2D algorithm works))
        if not os.path.exists(path):
            raise FileNotFoundError(f"The file you have entered does not exist: {path}")
        start = dt.datetime.strptime(start, fmt)

        reference_index = Event.create_reference_index(start=start, freq=freq)

        event_index = pd.read_csv(path, header=None)
        event_index.rename(columns={0: "id"}, inplace=True)
        # convert the index into date
        event_index = Event.ordinal_to_date(event_index, reference_index)
        # index difference maybe different from the duration as there might be a gap in the middle of the event.
        cols = ["continue", "ind_diff", "duration", "index"]
        event_index.loc[:, cols] = 0

        # the first day in the index file is an event beginning
        k = 1
        event_index.loc[0, "index"] = k
        event_beginning = event_index.loc[0, "date"]
        for i in range(1, len(event_index)):
            # if the day is previous day+1
            if (
                event_index.loc[i, "id"]
                <= event_index.loc[i - 1, "id"] + event_duration
            ):
                # then the event continues
                event_index.loc[i, "continue"] = 1
                # increase the duration
                event_index.loc[i, "ind_diff"] = event_index.loc[i - 1, "ind_diff"] + 1

                event_index.loc[i, "duration"] = (
                    event_index.loc[i, "date"] - event_beginning
                ).days + 1
            else:  # if not, then the day is the start of another event
                k = k + 1
                event_beginning = event_index.loc[i, "date"]

            event_index.loc[i, "index"] = k

        return cls(name, start=start, event_index=event_index)

    def get_all_events(self):
        """get_all_events.

            get_all_events method returns the end day of all events.

        Returns
        -------
            None.
        """
        assert hasattr(self, "event_index"), "please read/Create the event_index"
        ids = list()
        for i in range(len(self.event_index)):
            if self.event_index.loc[i, "continue"] == 0 and i != 0:
                ids.append(self.event_index.loc[i - 1, "id"])

        self.end_days = ids

    def create_from_overtopping(self, path: str, delimiter: str = r"\s+"):
        r"""create_from_overtopping.

            - Overtopping method reads the overtopping file and checks if the event_index dataframe has already been
            created by the CreateEventIndex method, it will add the overtopping to it, if not, it will create the
            event_index dataframe.

        Parameters
        ----------
        path: [str]
            path including the file name and extention of the Overtopping file result from the 1D model, the file has
            the following headers.
            >>>  Step	overtopping(m3/s)
            >>>  14342            850.8
            >>>  14373            893.4
            >>>  14374           1049.7
        delimiter: [str]
            Delimeter used in the overtopping file Default is space (r"\s+").

        Returns
        -------
        event_index: [DataFrame]
            this method creates an instance attribute of type
            dataframe with columns ['id','continue', 'ind_diff', 'duration',
            'Overtopping', 'OvertoppingCum', 'Volume']
        """
        overtop_total = pd.read_csv(path, delimiter=delimiter)
        # FIXME
        # if the flood event does not have overtopping for 1 day then continues to
        # overtop after the method considers it as two separate events however
        # it is the same event (if the gap is less than 10 days it is still
        # considered the same event)
        if not isinstance(self._event_index, DataFrame):
            # create the dataframe if the user did not use the CreateEventIndex method to
            # create the _event_index dataframe
            self._event_index = pd.DataFrame()
            self._event_index["id"] = overtop_total["Step"]
            self._event_index = self.ordinal_to_date(
                self._event_index, self.reference_index
            )

            self._event_index.loc[:, "continue"] = 0
            # index difference maybe different than the duration as there might be
            # a gap in the middle of the event
            self._event_index.loc[:, "ind_diff"] = 0
            self._event_index.loc[:, "duration"] = 0

            # the first day in the index file is an event beginning
            self.event_beginning = self._event_index.loc[0, "date"]
            for i in range(1, len(self._event_index)):
                # if the day is previous day+1
                if (
                    self._event_index.loc[i, "id"]
                    == self._event_index.loc[i - 1, "id"] + 1
                ):
                    # then the event continues
                    self._event_index.loc[i, "continue"] = 1
                    # increase the duration
                    self._event_index.loc[i, "ind_diff"] = (
                        self._event_index.loc[i - 1, "ind_diff"] + 1
                    )

                    self._event_index.loc[i, "duration"] = (
                        self._event_index.loc[i, "date"] - self.event_beginning
                    ).days + 1
                else:  # if not, then the day is the start of another event
                    self.event_beginning = self._event_index.loc[i, "date"]

        # store the overtopping data in the _event_index dataframe
        self._event_index["Overtopping"] = overtop_total["overtopping(m3/s)"]

        self._event_index.loc[0, "OvertoppingCum"] = self._event_index.loc[
            0, "Overtopping"
        ]
        for i in range(1, len(self._event_index)):
            if self._event_index.loc[i, "continue"] == 0:
                self._event_index.loc[i, "OvertoppingCum"] = self._event_index.loc[
                    i, "Overtopping"
                ]
            else:
                self._event_index.loc[i, "OvertoppingCum"] = (
                    self._event_index.loc[i, "Overtopping"]
                    + self._event_index.loc[i - 1, "OvertoppingCum"]
                )
        # the volume of water is m3/s for hourly stored and accumulated values
        # volume = overtopping * 60 *60 = m3
        self._event_index.loc[:, "Volume"] = (
            self._event_index.loc[:, "OvertoppingCum"] * 60 * 60
        )

    def calculate_volume_error(self, path):
        """calculate_volume_error.

            calculate_volume_error method reads the VoleError file, assign values to the coresponding time step.

        Parameters
        ----------
        path : [str]
            a path to the folder including the maps.

        Returns
        -------
        event_index: [dataframe].
            add columns ['DEMError','StepError','TooMuchWater'] to the event_index dataframe
        """
        # read the vol_error file
        vol_error = pd.read_csv(path, delimiter=r"\s+")
        self._event_index["DEMError"] = 0
        self._event_index["StepError"] = 0
        self._event_index["TooMuchWater"] = 0

        for i in range(len(vol_error)):
            loc = np.where(vol_error.loc[i, "step"] == self._event_index.loc[:, "id"])[
                0
            ][0]
            self._event_index.loc[
                loc, ["DEMError", "StepError", "TooMuchWater"]
            ] = vol_error.loc[
                i, ["DEM_Error", "PreviousDepthError", "TOOMuchWaterError"]
            ].tolist()

        self._event_index["VolError"] = (
            self._event_index["StepError"]
            + self._event_index["DEMError"]
            + self._event_index["TooMuchWater"]
        )
        self._event_index["VolError2"] = self._event_index["VolError"] / 20

    def overlay_maps(
        self,
        path: str,
        base_map: str,
        excluded_value,
    ) -> Catalog:
        """overlay_maps.

            - OverlayMaps method reads all the maps in the folder given by path input and overlays them with the
             basemap, and for each value in the basemap it creates a dictionary with the intersected values from all
             maps.

        Parameters
        ----------
        path: [String]
            a path to the folder including the maps.
        base_map: [String]
            a path including the name of the ASCII and extension like path="data/cropped.asc"
        excluded_value: [Numeric]
            values you want to exclude from extracted values.

        Returns
        -------
        Catalog:
            dictionary with a list of values in the basemap as keys and for each key a list of all the intersected
            values in the maps from the path.
        """
        classes_map = Dataset.read_file(base_map)
        # get the number of inundated cells in the Event index data frame.
        ind_unique = self.event_index.loc[:, "index"].unique()
        ind = self.event_index.loc[:, "index"].values
        event_catalog = {}

        for ind_i in ind_unique:
            event_catalog_i = {}
            row = np.where(ind == ind_i)[0][-1]
            # convert into int to be able to dump it to json
            event_i = self.event_index.loc[row, "id"]
            event_catalog_i["day"] = event_i
            inundation_map = f"{path}/{self.depth_prefix}{event_i}.tif"
            if os.path.exists(inundation_map):
                dataset = Dataset.read_file(inundation_map)
                depth_values = dataset.overlay(
                    classes_map,
                    excluded_value,
                )
                event_catalog_i["reaches"] = list(depth_values.keys())
                event_catalog_i["depth"] = depth_values
            else:
                event_catalog_i["reaches"] = []
                event_catalog_i["depth"] = {}

            event_catalog[event_i] = event_catalog_i

        catalog = Catalog(event_catalog)
        return catalog

    @classmethod
    def read_event_index(cls, name: str, path: str, start: str):
        """read_event_index.

            ReadEventIndex method reads the event_index table created using the "CreateEventIndex" or "Overtopping"
            methods.

        Parameters
        ----------
        name: [str]
            name of the river.
        path: [str]
            path of the index file result from the 2D model.
        start: [str]
            start date of the simulation.

        Returns
        -------
        event_index: [dataframe].
            dataframe of the event_index table
        """
        event_index = pd.read_csv(path)
        return cls(name, start=start, event_index=event_index)

    def histogram(
        self,
        day: int,
        exclude_value: Any = 0,
        map_type: int = 1,
        upper_bound=None,
        lower_bound=None,
        **kwargs,
    ):
        """histogram.

            - Histogram method extract values for the event MaxDepth map and plot the histogram the method checks
            first if you already extracted the values before then plot the histogram.

        Parameters
        ----------
        day: [Integer]
            DESCRIPTION.
        exclude_value: [Integer]
            DESCRIPTION.
        map_type: [integer], optional
            1 for the max depth maps, 2 for the duration map, 3 for the
            return period maps. The default is 1.
        upper_bound: [float, int]
            Default is 0.2
        lower_bound: [float, int]
            Default is 15
        """
        if hasattr(self, "extracted_values"):
            # get the list of event then extract their values.
            if day not in list(self.extracted_values.keys()):
                # depth map
                if map_type == 1:
                    path = f"{self.two_d_result_path}/{self.depth_prefix}{day}.tif"
                elif map_type == 2:
                    path = f"{self.two_d_result_path}/{self.duration_prefix}{day}.tif"
                else:
                    path = (
                        f"{self.two_d_result_path}/{self.return_period_prefix}{day}.tif"
                    )

                dataset = Dataset.read_file(path)
                extracted_values = dataset.extract(exclude_value=exclude_value)
                # non_zero_cells = len(extracted_values)
                self.extracted_values[day] = extracted_values

        extracted_values = self.extracted_values[day]
        # filter values
        if lower_bound is not None:
            extracted_values = [j for j in extracted_values if j > lower_bound]
        if upper_bound is not None:
            extracted_values = [j for j in extracted_values if j < upper_bound]

        hist = Statistic(extracted_values)
        ndims = extracted_values.ndim
        colors = ndims * ["#0504aa"]
        fig, ax, opts = hist.histogram(
            bins=15,
            color=colors,
            alpha=0.7,
            rwidth=0.85,
            ylabel="Frequency",
            **kwargs,
        )
        return fig, ax, opts

    def drop(self, drop_list):
        """drop.

            drop method deletes columns from the event_index dataframe.

        Parameters
        ----------
        drop_list: [list]
            list of column names to delete from the event_index dataframe table

        Returns
        -------
        event_index: [dataframe]
            the event_index dataframe without the columns in the drop_list
        """
        dataframe = self._event_index.loc[:, :]
        columns = list(dataframe.columns)

        [columns.remove(i) for i in drop_list]

        dataframe = dataframe.loc[:, columns]
        self._event_index = dataframe

    def to_csv(self, path: str):
        """Save.

            - Save method saves the event_index table.

        Parameters
        ----------
        path : [str]
            path to where you want to save the table.
        """
        self._event_index.to_csv(path, header=True, index=None)

    def get_event_start(self, loc: int) -> Tuple[int, int]:
        """get_event_start.

            - get_event_start method returns the index of the beginning of the event in the event_index dataframe.

        Parameters
        ----------
        loc: [int]
            dataframe index of the day you want to trace back to get the begining

        Returns
        -------
        ind: [int]
            dataframe index of the beginning day of the event

        Example
        ------
        if you want to get the beginning of the event that has the highest overtopping
        >>> HighOvertoppingInd = event_index['Overtopping'].idxmax()
        >>> ind = event_beginning(HighOvertoppingInd)
        """
        # loc = np.where(self.event_index['id'] == day)[0][0]
        # get all the days in the same event before that day as the inundation in the maps may
        # happen due to any of the days before not in this day
        ind = self.event_index.index[loc - self.event_index.loc[loc, "ind_diff"]]
        day = self.event_index.loc[ind, "id"]
        return ind, day

    def get_event_end(self, loc):
        """get_event_end.

            - method returns the index of the beginning of the event in the event_index dataframe.

        Parameters
        ----------
        loc: [int]
            dataframe index of the day you want to trace back to get the begining

        Returns
        -------
        ind: [int]
            dataframe index of the beginning day of the event

        Example
        -------
        if you want to get the beginning of the event that has the highest overtopping
        >>> HighOvertoppingInd = event_index['Overtopping'].idxmax()
        >>> ind = event_beginning(HighOvertoppingInd)
        """
        # loc = np.where(self.event_index['id'] == day)[0][0]
        # get all the days in the same event before that day as the inundation in the maps may
        # happen due to any of the days before not in this day

        # filter the dataframe and get only the 'indDiff' and 'id' columns
        filtered_event = self.event_index.loc[:, ["continue", "id"]]
        # get only days before the day you entered
        for i in range(loc + 1, len(filtered_event)):
            # start search from the following day
            if filtered_event.loc[i, "continue"] != 1:
                break

        ind = i - 1
        day = self.event_index.loc[ind, "id"]

        return ind, day

    def prepare_for_plotting(self, column_name: str):
        """prepare_for_plotting.

            - PrepareForPlotting takes a time series in the event_index dataframe and fill the days that do not
            exist in date column and fill it with zero to properly plot it without letting the graph mislead the
            viewer of connecting the data over the gap period.

        Parameters
        ----------
        column_name : [str]
            name of the columns you want.

        Returns
        -------
        DataFrame : [Dataframe]
            dataframe with a date column, and the required column
        """
        df = pd.DataFrame()
        df["date"] = self.reference_index["date"].tolist()
        df[column_name] = 0
        for i in range(len(self.event_index)):
            loc = np.where(df["date"] == self.event_index.loc[i, "date"])[0][0]
            df.loc[loc, column_name] = self.event_index.loc[i, column_name]

        return df



"""Cross-section model."""
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from statsmodels import api as sm


class CrossSections:
    """Cross-Section class."""

    def __init__(self, name):
        self.name = name

    def reg_plot(
        self,
        x,
        y,
        xlab,
        ylab,
        xlgd,
        ylgd,
        title,
        filename,
        log,
        logandlinear,
        seelim,
        Save=False,
        *args,
        **kwargs
    ):
        """reg_plot.

        Parameters
        ----------
        x : TYPE
            DESCRIPTION.
        y : TYPE
            DESCRIPTION.
        minmax_XS_area : TYPE
            DESCRIPTION.
        xlab : TYPE
            DESCRIPTION.
        ylab : TYPE
            DESCRIPTION.
        xlgd : TYPE
            DESCRIPTION.
        ylgd : TYPE
            DESCRIPTION.
        title : TYPE
            DESCRIPTION.
        filename : TYPE
            DESCRIPTION.
        log : [Bool]
            # Plot for log-log regression.
        logandlin : TYPE
            DESCRIPTION.
        seelim : [Bool]
            to draw vertical lines on the min and max drainage area on the graph.
        Save : TYPE, optional
            DESCRIPTION. The default is False.
        *args : TYPE
            DESCRIPTION.
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        None.


        Use the ordinary least squares to make a regression and plot the
        output.

        This function was developed first. The two following functions
        have been adapted from this one to meet some particular requirements.
        """
        for key in kwargs.keys():
            if key == "XLim":
                xmin = kwargs["XLim"][0]
                xmax = kwargs["XLim"][1]

        #    if log is True:
        #        # Linear regression with the ordinaty least squares method (Y, X).
        #        # sm.add_constant is required to get the intercept.
        #        # NaN values are dropped.
        #        results = sm.OLS(np.log10(y), sm.add_constant(np.log10(x)),
        #                         missing='drop').fit()
        #    elif log is False:
        #        results = sm.OLS(y, sm.add_constant(x), missing='drop').fit()
        # fit the relation between log(x) and log(y)
        results_log = sm.OLS(
            np.log10(y), sm.add_constant(np.log10(x)), missing="drop"
        ).fit()
        # fit the relation between x and y
        results_lin = sm.OLS(y, sm.add_constant(x), missing="drop").fit()
        # Print the results in the console
        #    print(results.summary())
        print(results_log.summary())
        print(results_lin.summary())

        # Retrieve the intercept and the slope
        intercept = results_lin.params[0]
        slope = results_lin.params[1]

        # Transform to log-type
        #    coeff = 10**intercept
        #    exp = slope
        coeff = 10 ** results_log.params[0]
        exp = results_log.params[1]

        # Transform to log-type if required
        #    if log is True:
        #        coeff = 10**intercept
        #        exp = slope

        # Retrieve r2
        rsq_log = results_log.rsquared
        rsq_lin = results_lin.rsquared

        # Save them to a datafram using an external function
        #    results_df = results_summary_to_dataframe(results)
        #    results_df.index = ['intercept_' + filename, 'slope_' + filename]
        #    results_df.to_csv(filename + '.csv', sep=',')

        if logandlinear is False:
            # logarithmic data
            results_log_df = self.results_summary_to_dataframe(results_log)
            results_log_df.index = ["intercept_" + filename, "slope_" + filename]
            if Save:
                results_log_df.to_csv(filename + "_powerlaw.csv", sep=",")
            # linear data
            results_lin_df = self.results_summary_to_dataframe(results_lin)
            results_lin_df.index = ["intercept_" + filename, "slope_" + filename]
            if Save:
                results_lin_df.to_csv(filename + "_lin.csv", sep=",")

        if logandlinear is False:
            # Plot the points and the regression line
            plt.scatter(x, y)
            x_plot = np.linspace(x.min(), x.max(), 1000)

            # Plot for log-log regression
            if log is True:
                plt.plot(x_plot, coeff * (x_plot**exp))
                plt.xscale("log")
                plt.yscale("log")
                plt.annotate(
                    "$%s = %.4f%s^{%.4f}$" % (ylgd, coeff, xlgd, exp)
                    + "\n"
                    + "$R^2 = %.4f$" % rsq_log,
                    xy=(0.05, 0.90),
                    xycoords="axes fraction",
                )
                plt.xlim(0.5 * x.min(), 1.5 * x.max())
                plt.ylim(0.5 * y.min(), 1.5 * y.max())
            # Plot for linear regression
            elif log is False:
                plt.plot(x_plot, intercept + slope * x_plot)
                plt.annotate(
                    "%s = %.4f + %.4f%s" % (ylgd, intercept, slope, xlgd)
                    + "\n"
                    + "$R^2 = %.4f$" % rsq_lin,
                    xy=(0.05, 0.90),
                    xycoords="axes fraction",
                )
                plt.xlim(0, x.max() + 0.1 * (x.max() - x.min()))
                plt.ylim(0, y.max() + 0.1 * (y.max() - y.min()))

            plt.xlabel(xlab)
            plt.ylabel(ylab)
            # vertical lines for min and max cross section area
            if seelim is True:
                plt.axvline(x=xmin, color="red", ymin=0.25, ymax=0.75)
                plt.axvline(x=xmax, color="red", ymin=0.25, ymax=0.75)
            plt.title(title)

            if Save:
                plt.savefig(filename + ".png", dpi=400)
                plt.close()
            if log is True:
                sm.graphics.plot_regress_exog(results_log, 1)
                if Save:
                    plt.savefig(filename + "_powerlaw_resid.png", dpi=400)
            elif log is False:
                sm.graphics.plot_regress_exog(results_lin, 1)
                if Save:
                    plt.savefig(filename + "_lin_resid.png", dpi=400)
                    plt.close()

        # Plot the power law and linear regressions on a log-log scale
        # and on a linear scale to compare the fits
        elif logandlinear is True:
            # Plot on a log-log scale
            plt.scatter(x, y, c="grey", marker="x")
            x_plot = np.linspace(x.min(), x.max(), 1000)
            plt.plot(x_plot, coeff * (x_plot**exp))
            plt.xscale("log")
            plt.yscale("log")
            plt.annotate(
                "$%s = %.4f%s^{%.4f}$" % (ylgd, coeff, xlgd, exp)
                + "\n"
                + "$R^2 = %.4f$" % rsq_log,
                xy=(0.05, 0.90),
                xycoords="axes fraction",
            )
            plt.plot(x_plot, intercept + slope * x_plot)
            plt.annotate(
                "$%s = %.4f + %.4f%s$" % (ylgd, intercept, slope, xlgd)
                + "\n"
                + "$R^2 = %.4f$" % rsq_lin,
                xy=(0.05, 0.80),
                xycoords="axes fraction",
            )
            plt.xlim(0.5 * x.min(), 1.5 * x.max())
            plt.ylim(0.5 * y.min(), 1.5 * y.max())
            plt.xlabel(xlab)
            plt.ylabel(ylab)

            if seelim is True:
                plt.axvline(x=xmin, color="red", ymin=0.25, ymax=0.75)
                plt.axvline(x=xmax, color="red", ymin=0.25, ymax=0.75)
            plt.title(title)
            if Save:
                plt.savefig(filename + "_logscale.png", dpi=400)
                plt.close()
            # Plot on a linear scale
            plt.scatter(x, y, c="grey", marker="x")
            x_plot = np.linspace(x.min(), x.max(), 1000)
            plt.plot(x_plot, coeff * (x_plot**exp))
            plt.annotate(
                "$%s = %.4f%s^{%.4f}$" % (ylgd, coeff, xlgd, exp)
                + "\n"
                + "$R^2 = %.4f$" % rsq_log,
                xy=(0.05, 0.90),
                xycoords="axes fraction",
            )
            plt.plot(x_plot, intercept + slope * x_plot)
            plt.annotate(
                "$%s = %.4f + %.4f%s$" % (ylgd, intercept, slope, xlgd)
                + "\n"
                + "$R^2 = %.4f$" % rsq_lin,
                xy=(0.05, 0.80),
                xycoords="axes fraction",
            )
            plt.xlim(0, x.max() + 0.1 * (x.max() - x.min()))
            plt.ylim(0, y.max() + 0.1 * (y.max() - y.min()))
            plt.xlabel(xlab)
            plt.ylabel(ylab)
            if seelim is True:
                plt.axvline(x=xmin, color="red", ymin=0.25, ymax=0.75)
                plt.axvline(x=xmax, color="red", ymin=0.25, ymax=0.75)
            plt.title(title)
            if Save:
                plt.savefig(filename + "_linscale.png", dpi=400)
                plt.close()

    def reg_plot_river(
        self, river_lst, data, minmax_XS_area, filename, log, Save=False, *args, **kargs
    ):
        """reg_plot_river.

        Use the ordinary least squares to make a regression and plot the
        output. This version makes use of the field 'river' in data to
        define a subset of gauges used for the regression.
        """
        # for key in kwargs.keys():
        #     if key == "XLim":
        #         xmin = kwargs['XLim'][0]
        #         xmax = kwargs['XLim'][1]
        #     if key == "area":
        #         minmax_XS_area = kwargs['area']

        df_output = pd.DataFrame()
        fig1 = plt.figure(figsize=(11.69, 16.53))
        for i, riveri in enumerate(river_lst):
            fig1.add_subplot(5, 2, i + 1)
            # The subset is defined checking the filed 'river' and using the
            # river names provided in river_lst
            datai = data[data["river"].str.contains(riveri, case=False) == True]
            # Warning, this df is passed outside of the function.
            min_XSUSarea = minmax_XS_area.at[riveri, "Min_UpXSArea"]
            max_XSUSarea = minmax_XS_area.at[riveri, "Max_UpXSArea"]
            print("There are {} gauges related to {}".format(datai.shape[0], riveri))
            # The regression is done if there are at least 4 gauges
            if datai.shape[0] > 3:
                x = datai["Area_Final"]
                y = datai["q2"]
                plt.scatter(x, y)
                X_plot = np.linspace(x.min(), x.max(), 1000)
                plt.ylabel("$HQ_2 [m^3/s]$")
                plt.xlabel("$Drainage Area [km^2]$")
                plt.axvline(x=min_XSUSarea, color="red", ymin=0.25, ymax=0.75)
                plt.axvline(x=max_XSUSarea, color="red", ymin=0.25, ymax=0.75)
                # Log-Log regression
                if log is True:
                    resultslog = sm.OLS(
                        np.log10(y),
                        sm.add_constant(np.log10(x)),  # Required to get the interc.
                        missing="drop",
                    ).fit()  # NaN values are dropped.
                    print(resultslog.summary())
                    results_df = self.results_summary_to_dataframe(resultslog)
                    results_df.index = ["interceptlog_" + riveri, "slopelog_" + riveri]
                    df_output = df_output.append(results_df)
                    coeffi = 10 ** resultslog.params[0]
                    expi = resultslog.params[1]
                    rsqi = resultslog.rsquared
                    plt.plot(
                        X_plot,
                        (10 ** resultslog.params[0]) * (X_plot ** resultslog.params[1]),
                    )
                    plt.xscale("log")
                    plt.yscale("log")
                    plt.annotate(
                        "$HQ_2 = %.3fA^{%.3f}$" % (coeffi, expi)
                        + "\n"
                        + "$R^2 = %.3f$" % rsqi,
                        xy=(0.05, 0.85),
                        xycoords="axes fraction",
                    )
                    plt.xlim(0.5 * x.min(), 1.5 * x.max())
                    plt.ylim(0.5 * y.min(), 1.5 * y.max())
                    plt.title(filename + " " + riveri + " log-log")
                # Linear regression
                elif log is False:
                    results = sm.OLS(
                        y,
                        sm.add_constant(x),  # Required to get the interc.
                        missing="drop",
                    ).fit()  # NaN values are dropped.
                    print(results.summary())
                    results_df = self.results_summary_to_dataframe(results)
                    results_df.index = [
                        "interceptlinear_" + riveri,
                        "slopelinear_" + riveri,
                    ]
                    df_output = df_output.append(results_df)
                    intercept = results.params[0]
                    slope = results.params[1]
                    rsq = results.rsquared
                    plt.plot(X_plot, intercept + slope * X_plot)
                    plt.annotate(
                        "Q2 = %.4f + %.4fA" % (intercept, slope)
                        + "\n"
                        + "$R^2 = %.4f$" % rsq,
                        xy=(0.05, 0.85),
                        xycoords="axes fraction",
                    )
                    plt.xlim(0, 1.1 * x.max())
                    plt.ylim(0, 1.1 * y.max())
                    plt.title(filename + " " + riveri + " linear")
                # plt.show()
        plt.tight_layout()
        # Plot the residuals.
        # Export for log-log regression
        if log is True:
            if Save:
                fig1.savefig(filename + "_LogLog.png", dpi=400)
            fig2 = sm.graphics.plot_regress_exog(resultslog, 1)
            if Save:
                fig2.savefig(filename + "_ResidLogLog.png", dpi=400)
                df_output.to_csv(filename + "_log.csv", sep=",")
        # Export for linear regression
        elif log is False:
            if Save:
                fig1.savefig(filename + "_Linear.png", dpi=400)
            fig2 = sm.graphics.plot_regress_exog(results, 1)
            if Save:
                fig2.savefig(filename + "_ResidLinear.png", dpi=400)
                df_output.to_csv(filename + "_linear.csv", sep=",")
        if Save:
            plt.close(fig1)
            plt.close(fig2)

    def reg_plot_subbasin(
        self, subbasin_lst, data, minmax_XS_area, filename, log, redfact, Save=False
    ):
        """reg_plot_subbasin Use the ordinary least squares to make a regression and plot the output.

        This version makes use of the field 'Subbasin' in data to define a subset
        of gauges used for the regression.

        redfact is a real between 0 and 1, and is used to define for each
        subbasin an area threshold that will limit the number of small gauges
        selected.
        """
        df_output = pd.DataFrame()
        fig1 = plt.figure(figsize=(11.69, 16.53))
        for i, subbasini in enumerate(subbasin_lst):
            fig1.add_subplot(5, 2, i + 1)
            datai = data[data["Subbasin"].str.contains(subbasini, case=False) == True]
            min_XSUSarea = minmax_XS_area.at[subbasini, "Min_UpXSArea"]
            min_gauge_area = redfact * min_XSUSarea
            max_XSUSarea = minmax_XS_area.at[subbasini, "Max_UpXSArea"]
            datai_red = datai[datai["Area_Final"] > min_gauge_area]
            print(
                "There are {} gauges used for {}".format(datai_red.shape[0], subbasini)
            )
            if datai.shape[0] > 3:
                x = datai_red["Area_Final"]
                y = datai_red["q2"]
                plt.scatter(x, y)
                X_plot = np.linspace(x.min(), x.max(), 1000)

                if log is True:
                    resultslog = sm.OLS(
                        np.log10(y),
                        sm.add_constant(np.log10(x)),  # Required to get the interc.
                        missing="drop",
                    ).fit()  # NaN values are dropped.
                    print(resultslog.summary())
                    results_df = self.results_summary_to_dataframe(resultslog)
                    results_df.index = [
                        "interceptlog_" + subbasini,
                        "slopelog_" + subbasini,
                    ]
                    df_output = df_output.append(results_df)
                    coeff = 10 ** resultslog.params[0]
                    exp = resultslog.params[1]
                    rsq = resultslog.rsquared
                    plt.plot(
                        X_plot,
                        (10 ** resultslog.params[0]) * (X_plot ** resultslog.params[1]),
                    )
                    plt.xscale("log")
                    plt.yscale("log")
                    plt.annotate(
                        "$Q2 = %.3fA^{%.3f}$" % (coeff, exp)
                        + "\n"
                        + "$R^2 = %.3f$" % rsq,
                        xy=(0.05, 0.85),
                        xycoords="axes fraction",
                    )
                    # plt.xlim(0.5*x.min(), 1.5*x.max())
                    plt.xlim(0.5 * min_gauge_area, 1.5 * max_XSUSarea)
                    plt.ylim(0.5 * y.min(), 1.5 * y.max())
                    plt.title(filename + " " + subbasini + " log-log")
                elif log is False:
                    results = sm.OLS(
                        y,
                        sm.add_constant(x),  # Required to get the interc.
                        missing="drop",
                    ).fit()  # NaN values are dropped.
                    print(results.summary())
                    results_df = self.results_summary_to_dataframe(results)
                    results_df.index = [
                        "interceptlinear_" + subbasini,
                        "slopelinear_" + subbasini,
                    ]
                    df_output = df_output.append(results_df)
                    intercept = results.params[0]
                    slope = results.params[1]
                    rsq = results.rsquared
                    plt.plot(X_plot, intercept + slope * X_plot)
                    plt.annotate(
                        "Q2 = %.4f + %.4fA" % (intercept, slope)
                        + "\n"
                        + "$R^2 = %.4f$" % rsq,
                        xy=(0.05, 0.85),
                        xycoords="axes fraction",
                    )
                    # plt.xlim(0.8*x.min(), 1.1*x.max())
                    plt.xlim(0, 1.1 * max_XSUSarea)
                    plt.ylim(0, 1.1 * y.max())
                    plt.title(filename + " " + subbasini + " linear")

                plt.ylabel("Q2 [m3/s]")
                plt.xlabel("Drainage Area [km2]")
                plt.axvline(x=min_XSUSarea, color="red", ymin=0.25, ymax=0.75)
                plt.axvline(x=max_XSUSarea, color="red", ymin=0.25, ymax=0.75)
                # plt.show()
        plt.tight_layout()
        if log is True:
            if Save:
                fig1.savefig(filename + "_LogLog.png", dpi=400)
                plt.close(fig1)
            fig2 = sm.graphics.plot_regress_exog(resultslog, 1)
            if Save:
                fig2.savefig(filename + "_ResidLogLog.png", dpi=400)
                df_output.to_csv(filename + "_LogLog.csv", sep=",")
        elif log is False:
            if Save:
                fig1.savefig(filename + "_Linear.png", dpi=400)
                plt.close(fig1)
            fig2 = sm.graphics.plot_regress_exog(results, 1)
            if Save:
                fig2.savefig(filename + "_ResidLinear.png", dpi=400)
                plt.close(fig2)
                df_output.to_csv(filename + "_Linear.csv", sep=",")

    @staticmethod
    def results_summary_to_dataframe(res):
        """Take the result of a statsmodel results table and transforms it into a dataframe."""

        pvals = res.pvalues
        coeff = res.params
        conf_lower = res.conf_int()[0]
        conf_higher = res.conf_int()[1]
        stderr = res.bse
        tvals = res.tvalues
        rsq = res.rsquared
        rsq_adj = res.rsquared_adj
        no_obs = res.nobs

        res_df = pd.DataFrame(
            {
                "pvals": pvals,
                "coeff": coeff,
                "conf_lower": conf_lower,
                "conf_higher": conf_higher,
                "std_err": stderr,
                "tvals": tvals,
                "R-squared": rsq,
                "R-squared_adj": rsq_adj,
                "no_obs": no_obs,
            }
        )

        # Reordering...
        res_df = res_df[
            [
                "coeff",
                "std_err",
                "tvals",
                "pvals",
                "conf_lower",
                "conf_higher",
                "R-squared",
                "R-squared_adj",
                "no_obs",
            ]
        ]
        return res_df

    def ListAttributes(self):
        """Print Attributes List."""

        print("\n")
        print(
            "Attributes List of: "
            + repr(self.__dict__["name"])
            + " - "
            + self.__class__.__name__
            + " Instance\n"
        )
        self_keys = list(self.__dict__.keys())
        self_keys.sort()
        for key in self_keys:
            if key != "name":
                print(str(key) + " : " + repr(self.__dict__[key]))

        print("\n")
